// 
//  Simple FIAP Uploader with NTP client for Sample Circuit (FIAP Sensor Node)
//
//  2011/09/19 ver.1   H.Inoue & H.Ochiai
//  2011/11/01 last updated
//
//  ★の部分はセンサ毎に設定（書き換え）が必要な項目

// 【説明】
// このプログラムは、トランジスタ技術誌2012年2月号に掲載されたIEEE1888関連の
// 記事のサンプルプログラムです。
// 記事中のセンサのハードウェアを制御し、測定したデータを用意したサーバに
// IEEE1888(FIAP)のプロトコルでアップロードを行う機能を持ちます。
// この版では、センサ自身の固定IPアドレス、ネットマスク、デフォルトゲートウェイ、
// またサーバのIPアドレスを、プログラム中に直に埋め込む必要があります。
// その反面、DHCPとDNSをサポートした版と比較して、メモリ領域（RAMエリア）に
// 余裕があります。
// プログラムの動作を理解するには、このプログラムが分かり易いと思います。
// 
//【コンパイルと実行の方法】
// １．次のライブラリを事前にArduino IDEのlibrariesフォルダへインストールしておきます
//   ・FIAPUploadAgentライブラリ
//   ・DateTimeライブラリ
//      http://www.arduino.cc/playground/Code/DateTime
// ２．このソースファイルでのローカル設定（センサごとの設定）の変更
//   ソースファイル中で ★ がついている変数や定数を、センサ毎に適切に設定して下さい
//   ・local_mac … MACアドレス（イーサネットアドレス）を設定します。
//        Ethernetシールドに MACアドレスが書かれている場合は、それを設定して下さい。
//        同じLAN内で重複があると通信ができなくなりますので注意下さい。
//   ・local_ip4, local_gateway, local_netmask … センサ自身のIPアドレスや
//        接続するネットワークの情報を適切に設定します。IPアドレスの重複に注意下さい。
//   ・server_ip4, server_host, server_path … データのアップロード先のサーバの
//        IPアドレスやホスト名とパスを設定します。ホスト名は実際には使用されず、
//        IPアドレスのほうが使用されますので、設定変更の際は注意して下さい。
//        アップロードのテストをするだけであれば、変更は必要ありません。
//   ・fiap_id_prefix … 必ずセンサ毎にサーバ上で一意となる文字列を設定してください。
//        httpで始まりますが、単なる文字列ですので、自分のメールアドレスやセンサの
//        設置場所の名称などを組み合わせてみてください。
//        文字列は、URLとして使える英数字・記号のみとなります。
//  ３．Arduino IDEでコンパイル＆ダウンロードを行います。
//  ４．Arduino IDEのSerial Monitorを起動すると、動作状況が表示されます。
//      また、LEDの状態（下記参照）から動作が正常かどうかを確認し、
//      サーバ上でアップロードしたデータが参照できるかどうかを確認してください。
// 
// 【RGB LEDの色の意味】
// ・白色 … NTPによる時刻情報の取得中
// ・青色 … サーバへのアクセスおよびアップロード処理中
// ・緑色 … アップロード成功
// ・消灯 … 次の測定までの待機中
// ・紫色 … サーバへの接続失敗（TCP接続失敗）
// ・黄色 … HTTP通信エラー
// ・水色 … IEEE1888通信エラー
// ・赤色 … その他のエラー


#include <SPI.h>              // Ethernetシールド用
#include <Ethernet.h>         // Ethernetシールド用
#include <FIAPUploadAgent.h>  // FIAPUploadAgent
#include <Time.h>             // 時刻関数・NTP用
#include <Udp.h>              // NTPパケット送受信用

// 全体パラメータ
const unsigned int MEASURE_INTERVAL = 60;  // アップロード間隔(秒)
char timezone[] = "+09:00";          // JST
const long timezoneOffset = +32400L; // 上記の時間を秒に直したもの（必ず同じにすること）
const byte LED_ON = 25;              // LED点灯時の明るさ(1-255)

// ローカル・ネットワーク設定
static byte local_mac[6] = { 0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED };  // ★ローカル環境に合わせる
static byte local_ip4[4]     = { 192, 168, 11, 55 }; // ★ローカル環境に合わせる
static byte local_gateway[4] = { 192, 168, 11, 1 };  // ★ローカル環境に合わせる
static byte local_netmask[4] = { 255, 255, 255, 0 }; // ★ローカル環境に合わせる

// IEEE1888サーバ設定
static byte server_ip4[4]  = { 133, 11, 168, 97 };                  // ★利用するサーバに合わせる
static char server_host[] = "fiap-sandbox.gutp.ic.i.u-tokyo.ac.jp"; // ★利用するサーバに合わせる
static char server_path[] = "/axis2/services/FIAPStorage";          // ★利用するサーバに合わせる
static int server_port = 80;

// PointSet ID (=PointIDのプレフィックス) の設定
static char fiap_id_prefix[] = "http://gutp.jp/Arduino/test-001/"; // ★他と重複しない自分のIDを使うこと

// temperature sensor port
#define TEMP_SENS   1
#define TEMP_COEF   10.0  // conversion parameter (mV per ℃)

// light intensity sensor port
#define LIGHT_SENS  0
#define LIGHT_COEF  0.35  // conversion parameter (mV per Lux)

// toggle SW port
#define SW_TOGGLE   16

// DIP SW ports
#define SW_DIP_1    4
#define SW_DIP_2    7
#define SW_DIP_3    8
#define SW_DIP_4    9

// RGB LED ports
#define LED_POLARITY  1	// 0: cathode common, 1: anode common
#define LED_R   3	// Red
#define LED_G   5	// Green
#define LED_B   6	// Blue

// センサ測定値保持フィールド
float temperature;
char str_temperature[7];
float illuminance;
char str_illuminance[7];
int dipsw;
char str_dipsw[5];
int tglsw;
char str_tglsw[5];

// 各ポイントの設定
struct fiap_element element[]=
{
  {"Temperature", str_temperature, 0,0,0,0,0,0,timezone},
  {"Illuminance", str_illuminance, 0,0,0,0,0,0,timezone},
  {"DIPSW", str_dipsw, 0,0,0,0,0,0,timezone},
  {"TGLSW", str_tglsw, 0,0,0,0,0,0,timezone},
};

// FIAPUploadAgent のインスタンス
FIAPUploadAgent FIAP;

// 初期設定ルーチン
void setup()
{
  int i;

  // I/O初期化
  Serial.begin(9600);
  Serial.println("Starting FIAP/IEEE1888 Uploader on Arduino");
  pinMode(TEMP_SENS, INPUT);
  pinMode(LIGHT_SENS, INPUT);
  pinMode(SW_TOGGLE, INPUT);
  pinMode(SW_DIP_1, INPUT);
  pinMode(SW_DIP_2, INPUT);
  pinMode(SW_DIP_3, INPUT);
  pinMode(SW_DIP_4, INPUT);
  pinMode(LED_R, OUTPUT);
  pinMode(LED_G, OUTPUT);
  pinMode(LED_B, OUTPUT);
  led_rgb_set(0, 0, 0);  // LED off
  
  // ネットワーク初期化
  Serial.print("IP = "); Serial.println(ip_to_str(local_ip4));
  Serial.print("NM = "); Serial.println(ip_to_str(local_netmask));
  Serial.print("GW = "); Serial.println(ip_to_str(local_gateway));
  Ethernet.begin(local_mac, local_ip4, local_gateway, local_netmask);

  // NTPサーバへの時刻問合せ
  Udp.begin(8888);    // for NTP client
  Serial.println("NTP querying ...");
  led_rgb_set(LED_ON, LED_ON, LED_ON);  // LED white
  setSyncProvider(NTPqueryTime);
  while (timeStatus() == timeNotSet) {
    time_t t = now();    // dummy call to NTP sync. process
  }
  led_rgb_set(0, 0, 0);  // LED off

  // FIAPライブラリの初期化
  FIAP.begin(server_ip4, server_host, server_path, server_port, fiap_id_prefix);
}

// メインループ
void loop()
{
  static long lcount = 0;  // loop counter
  int i;
  time_t t;

  // 各データに現在時刻をセット
  t = now();
  for(i = 0; i < 4; i++) {
     element[i].year = year(t);
     element[i].month = month(t);
     element[i].day = day(t);
     element[i].hour = hour(t);
     element[i].minute = minute(t);
     element[i].second = second(t);
  }

  // データ測定＆文字列として格納
  measure_data();
  dump_data();    // データをシリアルに出力

  // FIAPでアップロードを実行
  led_rgb_set(0, 0, LED_ON);  // LED blue
  Serial.print("FIAP uploading... ");
  int ret = FIAP.post(element, 4);
  Serial.print("status="); Serial.println(ret);
  switch (ret) {  // FIAPの結果に応じてLEDの色を点灯
  case FIAP_UPLOAD_OK:
    led_rgb_set(0, LED_ON, 0);  // LED green
    break;
  case FIAP_UPLOAD_CONNFAIL:
    led_rgb_set(LED_ON, 0, LED_ON);  // LED purple
    break;
  case FIAP_UPLOAD_HTTPERR:
    led_rgb_set(LED_ON, LED_ON, 0);  // LED yellow
    break;
  case FIAP_UPLOAD_FIAPERR:
    led_rgb_set(0, LED_ON, LED_ON);  // LED cyan
    break;
  default:
    led_rgb_set(LED_ON, 0, 0);  // LED red
    break;
  }

  // 測定間隔だけ待つ
  delay(MEASURE_INTERVAL * 500UL);
  led_rgb_set(0, 0, 0);  // LED off
  delay(MEASURE_INTERVAL * 500UL);  // 後半は結果のLEDを消す
}

// measure data and convert data to string
void measure_data()
{
  // 温度
  temperature = (float)analogRead(TEMP_SENS) * (5.0 / 1024.0) * 1000.0 / TEMP_COEF;
  sprintf(str_temperature, "%d.%1d", (int)temperature, (int)(temperature * 10) % 10);

  // 明るさ
  illuminance = (float)analogRead(LIGHT_SENS) * (5.0 / 1024.0) * 1000.0 / LIGHT_COEF;
  sprintf(str_illuminance, "%d", (int)illuminance);

  // DIP SW
  dipsw = (digitalRead(SW_DIP_4) == HIGH) * 8 + (digitalRead(SW_DIP_3) == HIGH) * 4
        + (digitalRead(SW_DIP_2) == HIGH) * 2 + (digitalRead(SW_DIP_1) == HIGH);
  sprintf(str_dipsw, "%d", dipsw);

  // Toggle SW
  tglsw = (digitalRead(SW_TOGGLE) == HIGH);
  if (tglsw) {
    sprintf(str_tglsw, "ON"); 
  } else {
    sprintf(str_tglsw, "OFF"); 
  }
}

// dump measured data
void dump_data()
{
  Serial.println();

  Serial.print("Date=");
  Serial.print((int)element[0].year);
  Serial.print("/");
  Serial.print((int)element[0].month);
  Serial.print("/");
  Serial.print((int)element[0].day); 
  Serial.print(" ");
  Serial.print((int)element[0].hour);
  Serial.print(":");
  Serial.print((int)element[0].minute);
  Serial.print(":");
  Serial.print((int)element[0].second); 
  Serial.print(" ");
  Serial.println(element[0].timezone);
  
  Serial.print("Temp=");
  Serial.println(str_temperature);
  Serial.print("Illuminance=");
  Serial.println(str_illuminance);
  Serial.print("DIP_SW=");
  Serial.println(str_dipsw);
  Serial.print("TOGGLE_SW=");
  Serial.println(str_tglsw);
}

// control RGB LED
void led_rgb_set(byte r, byte g, byte b)
{
  analogWrite(LED_R, LED_POLARITY ? (255 - (r)) : (r));
  analogWrite(LED_G, LED_POLARITY ? (255 - (g)) : (g));
  analogWrite(LED_B, LED_POLARITY ? (255 - (b)) : (b));
}

// IPアドレス(IPv4)をドットで区切られた文字列に変換する
char *ip_to_str(byte *ip)
{
  static char str[16];
  sprintf(str, "%d.%d.%d.%d\0", ip[0], ip[1], ip[2], ip[3]);
  return(str);
}

// Time adjustment via network using NTP(SNTP) 
// (This routine is based on UdpNtpClient program of Arduino Ethernet library)
const int NTP_MAX_WAITCOUNT = 10;    // wait count to receive response
const int NTP_PACKET_SIZE = 48;

// query NTP server and get current time
unsigned long NTPqueryTime()
{
  byte packetBuffer[NTP_PACKET_SIZE]; // NTP packet buffer
  byte ntpservers[][4] = {    // NTP server list
    { 133, 243, 238, 243 },   // ntp.nict.jp
    { 192, 43, 244, 18 }      // ntp.nist.gov
  };
  char n, i;
  
  // clear received packet before
  while (Udp.available()) { 
    Udp.readPacket(packetBuffer, NTP_PACKET_SIZE);
  }

  // query NTP servers
  for (n = 0; n < NTP_MAX_WAITCOUNT; n++) {
    for (i = 0; i < sizeof(ntpservers) / 4; i++) {
      sendNTPpacket(ntpservers[i], packetBuffer);
    }
    delay(500);
    if (!Udp.available()) { 
      continue;
    }
    memset(packetBuffer, 0, NTP_PACKET_SIZE); 
    Udp.readPacket(packetBuffer, NTP_PACKET_SIZE);

    unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
    unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
    unsigned long secsSince1900 = highWord << 16 | lowWord;
    if (secsSince1900 == 0) {
      continue;
    }
    unsigned long epoch = secsSince1900 - 2208988800UL + timezoneOffset;

    // NTP success, return time to time library
    Serial.print("NTP sync. time=");
    Serial.println(epoch);
    return(epoch);
  }

  // NTP failure
  Serial.println("NTP sync. failed");
  return(0);
}

// send an NTP request to the time server at the given address 
unsigned long sendNTPpacket(byte *address, byte *packetBuffer)
{
  memset(packetBuffer, 0, NTP_PACKET_SIZE); 
  packetBuffer[0]  = 0b11100011;
  packetBuffer[1]  = 0;
  packetBuffer[2]  = 6;
  packetBuffer[3]  = 0xEC;
  packetBuffer[12] = 49; 
  packetBuffer[13] = 0x4E;
  packetBuffer[14] = 49;
  packetBuffer[15] = 52;

  Udp.sendPacket(packetBuffer, NTP_PACKET_SIZE,  address, 123);
}

// end of code

